"""Unit tests for review-related actions."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from django.contrib.auth.models import User
from django.template import Context
from django.test.client import RequestFactory
from djblets.features.testing import override_feature_check
from djblets.siteconfig.models import SiteConfiguration
from djblets.testing.decorators import add_fixtures

from reviewboard.actions import actions_registry
from reviewboard.actions.base import AttachmentPoint
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.reviews.actions import (
    AddGeneralCommentAction,
    BaseReviewRequestAction,
    BaseReviewRequestMenuAction,
    CloseCompletedAction,
    CloseMenuAction,
    DeleteAction,
    DownloadDiffAction,
    EditReviewAction,
    LegacyAddGeneralCommentAction,
    LegacyEditReviewAction,
    LegacyShipItAction,
    ShipItAction,
    UpdateMenuAction,
    UploadDiffAction,
)
from reviewboard.reviews.features import unified_banner_feature
from reviewboard.reviews.models import ReviewRequest
from reviewboard.testing import TestCase


if TYPE_CHECKING:
    MixinParent = TestCase
else:
    MixinParent = object


class FooAction(BaseReviewRequestAction):
    action_id = 'foo-action'
    label = 'Foo Action'


class BarAction(BaseReviewRequestMenuAction):
    def __init__(self, action_id, child_actions=None):
        self.action_id = 'bar-' + action_id
        super().__init__(child_actions)


class TopLevelMenuAction(BaseReviewRequestMenuAction):
    action_id = 'top-level-menu-action'
    label = 'Top Level Menu Action'


class PoorlyCodedAction(BaseReviewRequestAction):
    def get_label(self, context):
        raise Exception


class ActionsTestCase(TestCase):
    """Test case for unit tests dealing with actions."""

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        actions_registry.populate()

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()
        actions_registry.reset()


class ReadOnlyActionTestsMixin(MixinParent):
    """Mixin for Review Board actions-related unit tests with read-only mode.

    This mixin is used to add read-only mode tests to action test cases.Using
    this mixin is especially important for actions that change visibility
    based on whether read-only mode is active. Actions that should always be
    visible can also be tested by setting ``read_only_always_show``.
    """

    uses_unified_banner = True

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        self.request = RequestFactory().request()
        self.siteconfig = SiteConfiguration.objects.get_current()

    def shortDescription(self) -> str:
        """Return an updated description for a particular test.

        If the test has an ``action`` attribute set and contains ``<ACTION>``
        in the docstring, then ACTION will be replaced by the ``action_id``
        attribute of the
        :py:class:`~reviewboard.reviews.actions.BaseReviewRequestAction`.

        Returns:
            str:
            The description of the test.
        """
        desc = super().shortDescription()

        if self.action and getattr(self, self._testMethodName, False):
            desc = desc.replace('<ACTION>', type(self.action).__name__)

        return desc

    def _create_request_context(self, *args, **kwargs) -> Context:
        """Create and return objects for use in the request context.

        Args:
            *args (tuple):
                Positional arguments for use in subclasses.

            **kwargs (dict):
                Keyword arguments for use in subclasses.

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context()

    def test_should_render_with_user_in_read_only(self) -> None:
        """Testing <ACTION>.should_render with authenticated user in read-only
        mode
        """
        self.request.user = User.objects.get(username='doc')

        # Turning on read-only mode prevents creation of some objects so call
        # _create_request_context first.
        request_context = self._create_request_context(user=self.request.user)

        settings = {
            'site_read_only': True,
        }

        with override_feature_check(unified_banner_feature.feature_id,
                                    self.uses_unified_banner):
            with self.siteconfig_settings(settings):
                if getattr(self, 'read_only_always_show', False):
                    self.assertTrue(
                        self.action.should_render(context=request_context))
                else:
                    self.assertFalse(
                        self.action.should_render(context=request_context))

    def test_should_render_with_superuser_in_read_only(self) -> None:
        """Testing <ACTION>.should_render with superuser in read-only mode"""
        self.request.user = User.objects.get(username='admin')

        # Turning on read-only mode prevents creation of some objects so call
        # _create_request_context first.
        request_context = self._create_request_context(user=self.request.user)

        settings = {
            'site_read_only': True,
        }

        with override_feature_check(unified_banner_feature.feature_id,
                                    self.uses_unified_banner):
            with self.siteconfig_settings(settings):
                self.assertTrue(
                    self.action.should_render(context=request_context))


class ActionRegistrationTests(ActionsTestCase):
    """Unit tests for legacy action registration.

    Deprecated:
        6.0:
        This can go away once we remove the legacy actions.
    """

    deprecation_message = (
        'BaseReviewRequestAction is deprecated and will be removed in '
        'Review Board 9. Please update your code to use '
        'reviewboard.actions.base.BaseAction')

    def test_action_register_methods(self) -> None:
        """Testing BaseReviewRequestAction.register and unregister"""
        with self.assertWarns(RemovedInReviewBoard90Warning,
                              self.deprecation_message):
            foo_action = FooAction()
            foo_action.register()

        self.assertEqual(actions_registry.get_action('foo-action'),
                         foo_action)

        foo_action.unregister()

        self.assertIsNone(actions_registry.get_action('foo-action'))

    def test_action_register_methods_with_parent(self) -> None:
        """Testing BaseReviewRequestAction.register and unregister with
        parent
        """
        with self.assertWarns(RemovedInReviewBoard90Warning,
                              self.deprecation_message):
            bar_action = BarAction('action-1')
            foo_action = FooAction()

        foo_placement = \
            foo_action.get_placement(AttachmentPoint.REVIEW_REQUEST)
        bar_placement = \
            bar_action.get_placement(AttachmentPoint.REVIEW_REQUEST)

        bar_action.register()
        foo_action.register(bar_action)

        self.assertIs(actions_registry.get_action('foo-action'), foo_action)
        self.assertIs(foo_placement.parent_action, bar_action)
        self.assertEqual(bar_placement.child_actions, [foo_action])

        foo_action.unregister()

        self.assertIsNone(actions_registry.get_action('foo-action'))
        self.assertIsNone(foo_placement.parent_action)
        self.assertEqual(bar_placement.child_actions, [])

    def test_menuaction_register_methods(self) -> None:
        """Testing BaseReviewRequestMenuAction.register and unregister"""
        with self.assertWarns(RemovedInReviewBoard90Warning,
                              self.deprecation_message):
            foo_action = FooAction()
            bar_action = BarAction('action-1', [foo_action])

        foo_placement = \
            foo_action.get_placement(AttachmentPoint.REVIEW_REQUEST)
        bar_placement = \
            bar_action.get_placement(AttachmentPoint.REVIEW_REQUEST)

        bar_action.register()

        self.assertEqual(actions_registry.get_action('foo-action'),
                         foo_action)
        self.assertEqual(actions_registry.get_action('bar-action-1'),
                         bar_action)
        self.assertIs(foo_placement.parent_action, bar_action)
        self.assertEqual(bar_placement.child_actions, [foo_action])

        bar_action.unregister()

        self.assertIsNone(actions_registry.get_action('foo-action'))
        self.assertIsNone(actions_registry.get_action('bar-action-1'))
        self.assertIsNone(foo_placement.parent_action)
        self.assertEqual(bar_placement.child_actions, [])

    def test_menuaction_register_methods_with_parent(self) -> None:
        """Testing BaseReviewRequestMenuAction.register and unregister with
        parent
        """
        with self.assertWarns(RemovedInReviewBoard90Warning,
                              self.deprecation_message):
            foo_action = FooAction()
            bar_action = BarAction('action-1', [foo_action])
            toplevel_action = TopLevelMenuAction()

        foo_placement = \
            foo_action.get_placement(AttachmentPoint.REVIEW_REQUEST)
        bar_placement = \
            bar_action.get_placement(AttachmentPoint.REVIEW_REQUEST)
        toplevel_placement = \
            toplevel_action.get_placement(AttachmentPoint.REVIEW_REQUEST)

        toplevel_action.register()

        bar_action.register(toplevel_action)

        self.assertIs(actions_registry.get_action('foo-action'), foo_action)
        self.assertIs(actions_registry.get_action('bar-action-1'), bar_action)
        self.assertEqual(toplevel_placement.child_actions, [bar_action])
        self.assertIs(bar_placement.parent_action, toplevel_action)
        self.assertIs(foo_placement.parent_action, bar_action)
        self.assertEqual(bar_placement.child_actions, [foo_action])

        bar_action.unregister()

        self.assertIsNone(actions_registry.get_action('foo-action'))
        self.assertIsNone(actions_registry.get_action('bar-action-1'))
        self.assertIsNone(foo_placement.parent_action)
        self.assertEqual(bar_placement.child_actions, [])
        self.assertEqual(toplevel_placement.child_actions, [])

        self.assertIsNone(bar_placement.parent_action)


class AddGeneralCommentActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for AddGeneralCommentAction."""

    action = AddGeneralCommentAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        user: Optional[User] = None,
        url_name: str = 'review-request-detail',
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to run the request.

            url_name (str, optional):
                The URL name to fake on the resolver.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(user=user, url_name=url_name),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing AddGeneralCommentAction.should_render with authenticated
        user
        """
        user = User.objects.get(username='doc')
        self.request.user = user

        self.assertTrue(self.action.should_render(
            context=self._create_request_context(user)))

    def test_should_render_with_anonymous(self) -> None:
        """Testing AddGeneralCommentAction.should_render with authenticated
        user
        """
        self.assertFalse(
            self.action.should_render(context=self._create_request_context()))


class LegacyAddGeneralCommentActionTests(ReadOnlyActionTestsMixin,
                                         ActionsTestCase):
    """Unit tests for LegacyAddGeneralCommentAction."""

    action = LegacyAddGeneralCommentAction()
    fixtures = ['test_users']
    uses_unified_banner = False

    def _create_request_context(
        self,
        user: Optional[User] = None,
        url_name: str = 'review-request-detail',
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to run the request.

            url_name (str, optional):
                The URL name to fake on the resolver.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(user=user, url_name=url_name),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing LegacyAddGeneralCommentAction.should_render with
        authenticated user
        """
        user = User.objects.get(username='doc')

        with override_feature_check(unified_banner_feature.feature_id, False):
            self.request.user = user
            self.assertTrue(self.action.should_render(
                context=self._create_request_context(user)))

    def test_should_render_with_anonymous(self) -> None:
        """Testing LegacyAddGeneralCommentAction.should_render with
        authenticated user
        """
        with override_feature_check(unified_banner_feature.feature_id, False):
            self.assertFalse(
                self.action.should_render(context=self._create_request_context()))


class CloseMenuActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for CloseMenuAction."""

    action = CloseMenuAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        can_change_status: bool = True,
        public: bool = True,
        status: str = ReviewRequest.PENDING_REVIEW,
        user: Optional[User] = None,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            can_change_status (bool, optional):
                Whether the ``can_change_status`` permission should be set.

            public (bool, optional):
                Whether the review request should be public.

            status (str, optional):
                The status for the review request.

            user (django.contrib.auth.models.User, optional):
                An optional user to set as the owner of the request.

            *args (tuple):
                Additional positional arguments (unused).

            **kwargs (dict):
                Additional keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        review_request = self.create_review_request(
            public=public, status=status)
        request = self.create_http_request(
            user=user or review_request.submitter,
            url_name=url_name)

        return Context({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': can_change_status,
                },
            },
        })

    def test_should_render_for_owner(self) -> None:
        """Testing CloseMenuAction.should_render for owner of review request"""
        self.assertTrue(self.action.should_render(
            context=self._create_request_context()))

    def test_should_render_for_owner_unpublished(self) -> None:
        """Testing CloseMenuAction.should_render for owner of review
        unpublished review request
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                public=False)))

    def test_should_render_for_user(self) -> None:
        """Testing CloseMenuAction.should_render for normal user"""
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                can_change_status=False,
                user=self.create_user())))

    def test_should_render_user_with_can_change_status(self) -> None:
        """Testing CloseMenuAction.should_render for user with
        can_change_status permission
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                can_change_status=True,
                user=self.create_user())))

    def test_should_render_user_with_can_change_status_and_unpublished(
        self,
    ) -> None:
        """Testing CloseMenuAction.should_render for user with
        can_change_status permission and unpublished review request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                can_change_status=True,
                public=False,
                user=self.create_user())))

    def test_should_render_with_discarded(self) -> None:
        """Testing CloseMenuAction.should_render with discarded review request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                status=ReviewRequest.DISCARDED)))

    def test_should_render_with_submitted(self) -> None:
        """Testing CloseMenuAction.should_render with submitted review request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                status=ReviewRequest.SUBMITTED)))


class DeleteActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for DeleteAction."""

    fixtures = ['test_users']
    action = DeleteAction()

    def _create_request_context(
        self,
        user: Optional[User] = None,
        url_name: str = 'review-request-detail',
        delete_reviewrequest: bool = True,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to run the request.

            url_name (str, optional):
                The URL name to fake on the resolver.

            delete_reviewrequest (bool, optional):
                Whether the resulting context should include the
                ``delete_reviewrequest`` permission.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(user=user, url_name=url_name),
            'perms': {
                'reviews': {
                    'delete_reviewrequest': delete_reviewrequest,
                },
            },
        })

    def test_should_render_with_published(self) -> None:
        """Testing DeleteAction.should_render with standard user"""
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                user=self.create_user(),
                delete_reviewrequest=False)))

    def test_should_render_with_permission(self) -> None:
        """Testing DeleteAction.should_render with delete_reviewrequest
        permission
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                user=self.create_user())))


class DownloadDiffActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for DownloadDiffAction."""

    action = DownloadDiffAction()
    fixtures = ['test_users']
    read_only_always_show = True

    def _create_request_context(
        self,
        review_request: Optional[ReviewRequest] = None,
        url_name: str = 'view-diff',
        with_local_site: bool = False,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest, optional):
                The review request to use. If not specified, one will be
                created.

            url_name (str, optional):
                The URL name to fake on the resolver.

            with_local_site (bool, optional):
                Whether to use a local site.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        if not review_request:
            review_request = self.create_review_request(
                with_local_site=with_local_site)

        return Context({
            'request': self.create_http_request(
                user=self.create_user(),
                with_local_site=with_local_site,
                url_name=url_name),
            'review_request': review_request,
        })

    def test_get_url_on_diff_viewer(self) -> None:
        """Testing DownloadDiffAction.get_url on diff viewer page"""
        self.assertEqual(
            self.action.get_url(context=self._create_request_context()),
            'raw/')

    def test_get_url_on_interdiff(self) -> None:
        """Testing DownloadDiffAction.get_url on diff viewer interdiff page"""
        self.assertEqual(
            self.action.get_url(context=self._create_request_context(
                url_name='view-interdiff')),
            'raw/')

    def test_get_url_on_diff_viewer_revision(self) -> None:
        """Testing DownloadDiffAction.get_url on diff viewer revision page"""
        self.assertEqual(
            self.action.get_url(context=self._create_request_context(
                url_name='view-diff-revision')),
            'raw/')

    def test_get_url_on_review_request(self) -> None:
        """Testing DownloadDiffAction.get_url on review request page"""
        review_request = self.create_review_request()

        self.assertEqual(
            self.action.get_url(context=self._create_request_context(
                review_request=review_request,
                url_name='review-request-detail')),
            '/r/%s/diff/raw/' % review_request.display_id)

    @add_fixtures(['test_site'])
    def test_get_url_on_review_request_with_local_site(self) -> None:
        """Testing DownloadDiffAction.get_url on review request page with
        LocalSite
        """
        review_request = self.create_review_request(id=123,
                                                    with_local_site=True)

        self.assertEqual(
            self.action.get_url(context=self._create_request_context(
                review_request=review_request,
                url_name='review-request-detail',
                with_local_site=True)),
            '/s/%s/r/%s/diff/raw/' % (self.local_site_name,
                                      review_request.display_id))

    def test_get_hidden_on_diff_viewer(self) -> None:
        """Testing DownloadDiffAction.get_visible on diff viewer page"""
        self.assertTrue(self.action.get_visible(
            context=self._create_request_context(url_name='view-diff')))

    def test_get_hidden_on_interdiff(self) -> None:
        """Testing DownloadDiffAction.get_visible on diff viewer interdiff page
        """
        self.assertFalse(self.action.get_visible(
            context=self._create_request_context(url_name='view-interdiff')))

    def test_get_hidden_on_diff_viewer_revision(self) -> None:
        """Testing DownloadDiffAction.get_visible on diff viewer revision page
        """
        self.assertTrue(self.action.get_visible(
            context=self._create_request_context(
                url_name='view-diff-revision')))

    def test_get_hidden_on_review_request(self) -> None:
        """Testing DownloadDiffAction.get_visible on diff viewer revision page
        """
        self.assertTrue(self.action.get_visible(
            context=self._create_request_context(
                url_name='review-request-detail')))

    def test_should_render_on_diff_viewer(self) -> None:
        """Testing DownloadDiffAction.should_render on diff viewer page"""
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                url_name='view-diff')))

    def test_should_render_on_interdiff(self) -> None:
        """Testing DownloadDiffAction.should_render on diff viewer interdiff
        page
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                url_name='view-diff-revision')))

    def test_should_render_on_diff_viewer_revision(self) -> None:
        """Testing DownloadDiffAction.should_render on diff viewer revision
        page
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                url_name='view-diff-revision')))

    @add_fixtures(['test_scmtools'])
    def test_should_render_on_review_request_with_repository(self) -> None:
        """Testing DownloadDiffAction.should_render on review request page
        with repository but no diff
        """
        review_request = self.create_review_request(create_repository=True)

        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                review_request=review_request,
                url_name='review-request-detail')))

    @add_fixtures(['test_scmtools'])
    def test_should_render_on_review_request_with_repository_and_diff(
        self,
    ) -> None:
        """Testing DownloadDiffAction.should_render on review request page
        with repository and diff history
        """
        review_request = self.create_review_request(create_repository=True)
        self.create_diffset(review_request)

        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                review_request=review_request,
                url_name='review-request-detail')))

    @add_fixtures(['test_scmtools'])
    def test_should_render_on_review_request_without_repository(self) -> None:
        """Testing DownloadDiffAction.should_render on review request page
        without repository
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                url_name='review-request-detail')))


class EditReviewActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for EditReviewAction."""

    action = EditReviewAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            user (django.contrib.auth.models.User, optional):
                The user to set on the request.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(url_name=url_name,
                                                user=user),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing EditReviewAction.should_render with authenticated user
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                user=User.objects.get(username='doc'))))

    def test_should_render_with_anonymous(self) -> None:
        """Testing EditReviewAction.should_render with authenticated user
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context()))


class LegacyEditReviewActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for LegacyEditReviewAction."""

    action = LegacyEditReviewAction()
    fixtures = ['test_users']
    uses_unified_banner = False

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            user (django.contrib.auth.models.User, optional):
                The user to set on the request.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(url_name=url_name,
                                                user=user),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing LegacyEditReviewAction.should_render with authenticated
        user
        """
        with override_feature_check(unified_banner_feature.feature_id, False):
            self.assertTrue(self.action.should_render(
                context=self._create_request_context(
                    user=User.objects.get(username='doc'))))

    def test_should_render_with_anonymous(self) -> None:
        """Testing LegacyEditReviewAction.should_render with authenticated
        user
        """
        with override_feature_check(unified_banner_feature.feature_id, False):
            self.assertFalse(self.action.should_render(
                context=self._create_request_context()))

    def test_should_render_with_user_in_read_only(self) -> None:
        """Testing LegacyEditReviewAction.should_render with authenticated
        user in read-only mode
        """
        with override_feature_check(unified_banner_feature.feature_id, False):
            super().test_should_render_with_user_in_read_only()

    def test_should_render_with_superuser_in_read_only(self) -> None:
        """Testing LegacyEditReviewAction.should_render with superuser in
        read-only mode
        """
        with override_feature_check(unified_banner_feature.feature_id, False):
            super().test_should_render_with_superuser_in_read_only()


class ShipItActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for ShipItAction."""

    action = ShipItAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            user (django.contrib.auth.models.User, optional):
                The user to set on the request.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(url_name=url_name,
                                                user=user),
            'review_request': self.create_review_request(submitter='grumpy'),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing ShipItAction.should_render with authenticated user"""
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                user=User.objects.get(username='doc'))))

    def test_should_render_with_anonymous(self) -> None:
        """Testing ShipItAction.should_render with anonymous user"""
        self.assertFalse(self.action.should_render(
            context=self._create_request_context()))

    def test_should_render_with_no_self_ship_it(self) -> None:
        """Testing ShipItAction.should_render with a user reviewing their own
        review request and reviews_allow_self_shipit=False
        """
        with self.siteconfig_settings({'reviews_allow_self_shipit': False}):
            self.assertFalse(self.action.should_render(
                context=self._create_request_context(
                    user=User.objects.get(username='grumpy'))))

    def test_should_render_with_self_ship_it(self) -> None:
        """Testing ShipItAction.should_render with a user reviewing their own
        review request and reviews_allow_self_shipit=True
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                user=User.objects.get(username='grumpy'))))


class LegacyShipItActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for LegacyShipItAction."""

    action = LegacyShipItAction()
    fixtures = ['test_users']
    uses_unified_banner = False

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            user (django.contrib.auth.models.User, optional):
                The user to set on the request.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        return Context({
            'request': self.create_http_request(url_name=url_name,
                                                user=user),
            'review_request': self.create_review_request(submitter='grumpy'),
        })

    def test_should_render_with_authenticated(self) -> None:
        """Testing ShipItAction.should_render with authenticated user"""
        with override_feature_check(unified_banner_feature.feature_id, False):
            self.assertTrue(self.action.should_render(
                context=self._create_request_context(
                    user=User.objects.get(username='doc'))))

    def test_should_render_with_anonymous(self) -> None:
        """Testing ShipItAction.should_render with anonymous user"""
        with override_feature_check(unified_banner_feature.feature_id, False):
            self.assertFalse(self.action.should_render(
                context=self._create_request_context()))


class CloseCompletedActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for CloseCompletedAction."""

    action = CloseCompletedAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        public: bool = True,
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            public (bool, optional):
                Whether the review request should be public.

            user (django.contrib.auth.models.User, optional):
                The user to check visibility for.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        review_request = self.create_review_request(public=public)

        return Context({
            'request': self.create_http_request(
                user=user or review_request.submitter,
                url_name=url_name),
            'review_request': review_request,
        })

    def test_should_render_with_published(self) -> None:
        """Testing CloseCompletedAction.should_render with published
        review request
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(public=True)))

    def test_should_render_with_unpublished(self) -> None:
        """Testing CloseCompletedAction.should_render with unpublished review
        request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(public=False)))


class UpdateMenuActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for UpdateMenuAction."""

    action = UpdateMenuAction()
    fixtures = ['test_users']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        public: bool = True,
        status: str = ReviewRequest.PENDING_REVIEW,
        user: Optional[User] = None,
        can_edit_reviewrequest: bool = True,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            public (bool, optional):
                Whether the review request should be public.

            status (str, optional):
                Review request status.

            user (django.contrib.auth.models.User, optional):
                The user to check visibility for.

            can_edit_reviewrequest (bool, optional):
                Whether the ``can_edit_reviewrequest`` permission should be
                set.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        review_request = self.create_review_request(public=public,
                                                    status=status)

        return Context({
            'review_request': review_request,
            'request': self.create_http_request(
                user=user or review_request.submitter,
                url_name=url_name),
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': can_edit_reviewrequest,
                },
            },
        })

    def test_should_render_for_owner(self) -> None:
        """Testing UpdateMenuAction.should_render for owner of review request
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(
                can_edit_reviewrequest=False)))

    def test_should_render_for_user(self) -> None:
        """Testing UpdateMenuAction.should_render for normal user"""
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                user=self.create_user(),
                can_edit_reviewrequest=False)))

    def test_should_render_user_with_can_edit_reviewrequest(self) -> None:
        """Testing UpdateMenuAction.should_render for user with
        can_edit_reviewrequest permission
        """
        self.assertTrue(self.action.should_render(
            context=self._create_request_context(user=self.create_user())))

    def test_should_render_with_discarded(self) -> None:
        """Testing UpdateMenuAction.should_render with discarded review request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                status=ReviewRequest.DISCARDED,
                can_edit_reviewrequest=False)))

    def test_should_render_with_submitted(self) -> None:
        """Testing UpdateMenuAction.should_render with submitted review request
        """
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(
                status=ReviewRequest.SUBMITTED,
                can_edit_reviewrequest=False)))


class UploadDiffActionTests(ReadOnlyActionTestsMixin, ActionsTestCase):
    """Unit tests for UploadDiffAction."""

    action = UploadDiffAction()
    fixtures = ['test_users', 'test_scmtools']

    def _create_request_context(
        self,
        url_name: str = 'review-request-detail',
        can_edit_reviewrequest: bool = True,
        create_repository: bool = True,
        user: Optional[User] = None,
        *args,
        **kwargs,
    ) -> Context:
        """Create and return objects for use in the request context.

        Args:
            url_name (str, optional):
                The URL name to fake on the resolver.

            create_repository (bool, optional):
                Whether to create a repository for the review request.

            user (django.contrib.auth.models.User, optional):
                The user to check visibility for.

            *args (tuple):
                Positional arguments (unused).

            **kwargs (dict):
                Keyword arguments (unused).

        Returns:
            django.template.Context:
            Additional context to use when testing read-only actions.
        """
        review_request = self.create_review_request(
            create_repository=create_repository)

        return Context({
            'review_request': review_request,
            'request': self.create_http_request(
                user=user or review_request.submitter,
                url_name=url_name),
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': can_edit_reviewrequest,
                },
            },
        })

    def test_get_label_with_no_diffs(self) -> None:
        """Testing UploadDiffAction.get_label with no diffs"""
        review_request = self.create_review_request()
        self.request.user = review_request.submitter

        self.assertEqual(
            self.action.get_label(context=Context({
                'review_request': review_request,
                'request': self.request,
            })),
            'Upload Diff')

    def test_get_label_with_diffs(self) -> None:
        """Testing UploadDiffAction.get_label with diffs"""
        review_request = self.create_review_request(create_repository=True)
        self.create_diffset(review_request)

        self.request.user = review_request.submitter

        self.assertEqual(
            self.action.get_label(context=Context({
                'review_request': review_request,
                'request': self.request,
            })),
            'Update Diff')

    def test_should_render_with_repository(self) -> None:
        """Testing UploadDiffAction.should_render with repository"""
        self.assertTrue(self.action.should_render(
            context=self._create_request_context()))

    def test_should_render_without_repository(self) -> None:
        """Testing UploadDiffAction.should_render without repository"""
        self.assertFalse(self.action.should_render(
            context=self._create_request_context(create_repository=False)))



"""Utility methods for read-only mode."""

from djblets.siteconfig.models import SiteConfiguration


def is_site_read_only_for(user):
    """Check whether user should be affected by read-only mode.

    Superusers are not affected by read-only mode. Otherwise, check the current
    :py:class:`~djblets.siteconfig.models.SiteConfiguration` for
    ``site_read_only``.

    Args:
        user (django.contrib.auth.models.User):
            The user that is to be checked.

    Returns:
        bool:
        Whether the site is read-only for a user.
    """
    if user and user.is_superuser:
        return False

    if user and hasattr(user, '_is_site_read_only'):
        return user._is_site_read_only

    siteconfig = SiteConfiguration.objects.get_current()
    is_read_only = siteconfig.get('site_read_only')

    if user:
        user._is_site_read_only = is_read_only

    return is_read_only



"""Hosting service for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING
from urllib.parse import quote as urlquote, urljoin, urlparse

from django.template.loader import render_to_string
from django.urls import path
from django.utils.translation import gettext_lazy as _

from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.errors import HostingServiceError
from reviewboard.hostingsvcs.forgejo.client import ForgejoClient
from reviewboard.hostingsvcs.forgejo.errors import APITokenNameInUseError
from reviewboard.hostingsvcs.forgejo.forms import ForgejoForm
from reviewboard.hostingsvcs.forgejo.views import WebHookView
from reviewboard.scmtools.crypto_utils import encrypt_password
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django.http import HttpRequest
    from django.utils.safestring import SafeString

    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.scmtools.core import Branch, Commit
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class Forgejo(BaseHostingService, BaseBugTracker):
    """Hosting service for Forgejo.

    Version Added:
        8.0
    """

    hosting_service_id = 'forgejo'
    name = _('Forgejo')

    client_class = ForgejoClient
    form = ForgejoForm

    has_repository_hook_instructions = True
    needs_authorization = True
    self_hosted = True
    supported_scmtools = ['Git']
    supports_bug_trackers = True
    supports_post_commit = True
    supports_repositories = True
    supports_two_factor_auth = True

    bug_tracker_field = \
        '%(hosting_url)s/%(repository_owner)s/%(repository_name)s/issues/%%s'
    repository_fields = {
        'Git': {
            'path': '%(hosting_url)s/%(repository_owner)s/'
                    '%(repository_name)s.git',
        },
    }

    repository_url_patterns = [
        path('hooks/close-submitted/',
             view=WebHookView.as_view(),
             name='forgejo-hooks-close-submitted'),
    ]

    # Set this for typing.
    client: ForgejoClient

    def is_authorized(self) -> bool:
        """Return whether the account has stored authorization.

        This will check to see if we have stored an API token for the account.
        It does not validate that the token still works.

        Returns:
            bool:
            ``True`` if the account has a saved API token. ``False``,
            otherwise.
        """
        return 'api_token' in self.account.data

    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        hosting_url: str | None,
        two_factor_auth_code: (str | None) = None,
        **kwargs,
    ) -> None:
        """Authorize an account for Forgejo.

        Args:
            username (str):
                The username for the account.

            password (str):
                The password for the account.

            hosting_url (str):
                The URL of the Forgejo server.

            two_factor_auth_code (str, optional):
                The two-factor authentication code provided by the user.

            **kwargs (dict, unused):
                Extra keyword arguments containing values from the
                repository's configuration.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.TwoFactorAuthCodeRequiredError:
                A two-factor authentication code is required to authorize
                this account. The request must be retried with the same
                credentials and with the ``two_factor_auth_code`` parameter
                provided.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                A non-authentication error occurred.
        """
        assert username is not None
        assert password is not None
        assert hosting_url is not None

        attempts = 0
        max_tries = 5

        server_name = urlparse(get_server_url()).hostname
        assert server_name is not None

        while attempts < max_tries:
            token_name = self._create_api_token_name(server_name)

            try:
                api_token = self.client.create_api_token(
                    token_name=token_name,
                    hosting_url=hosting_url,
                    username=username,
                    password=password,
                    two_factor_auth_code=two_factor_auth_code)
            except APITokenNameInUseError:
                attempts += 1
                continue

            break
        else:
            logger.error('Failed to create an API token with a unique name '
                         'for user %s on server %s after %d attempts.',
                         username, hosting_url, max_tries)

            raise HostingServiceError(
                _(
                    'Unable to create a Forgejo API token with a unique name '
                    'after {tries} attempts for user {username} on server '
                    '{hosting_url}'
                ).format(
                    tries=max_tries,
                    username=username,
                    hosting_url=hosting_url,
                )
            )

        self.account.data['api_token'] = encrypt_password(api_token)
        self.account.save()

    def check_repository(
        self,
        *,
        repository_owner: str,
        repository_name: str,
        **kwargs,
    ) -> None:
        """Check the validity of a repository configuration.

        This checks that the data provided in the repository form is valid.

        Args:
            repository_owner (str):
                The repository owner.

            repository_name (str):
                The name of the repository.

            **kwargs (dict):
                Additional keyword arguments from the repository form.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository is not valid.
        """
        self.client.check_repository(
            hosting_url=self.account.hosting_url,
            repository_owner=repository_owner,
            repository_name=repository_name)

    def get_file(
        self,
        repository: Repository,
        path: str,
        revision: str,
        *args,
        **kwargs,
    ) -> bytes:
        """Return the requested file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (str):
                The file path.

            revision (str):
                The revision the file should be retrieved from.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found or the API could not be accessed.
        """
        return self.client.get_blob(
            hosting_url=self.account.hosting_url,
            repository=repository,
            path=path,
            sha=revision)

    def get_branches(
        self,
        repository: Repository,
    ) -> Sequence[Branch]:
        """Return a list of all branches in the repositories.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository for which branches should be returned.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching branches.
        """
        return self.client.get_branches(
            hosting_url=self.account.hosting_url,
            repository=repository)

    def get_commits(
        self,
        repository: Repository,
        branch: (str | None) = None,
        start: (str | None) = None,
    ) -> Sequence[Commit]:
        """Return a list of commits backward in history from a given point.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve commits from.

            branch (str, optional):
                The branch to retrieve from.

                If this is not provided, the default branch will be used.

            start (str, optional):
                An optional starting revision.

                If this is not provided, the most recent commits will be
                returned.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The retrieved commits.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching commits.
        """
        return self.client.get_commits(
            hosting_url=self.account.hosting_url,
            repository=repository,
            branch=branch,
            start=start)

    def get_change(
        self,
        repository: Repository,
        revision: str,
    ) -> Commit:
        """Return an individual change.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to get the change from.

            revision (str):
                The revision to retrieve.

        Returns:
            reviewboard.scmtools.core.Commit:
            The change.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching the commit.
        """
        return self.client.get_change(
            hosting_url=self.account.hosting_url,
            repository=repository,
            revision=revision)

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.bugtracker.BugInfo:
            Information about the bug.
        """
        return self.client.get_bug_info(
            hosting_url=self.account.hosting_url,
            repository=repository,
            bug_id=bug_id)

    def get_repository_hook_instructions(
        self,
        request: HttpRequest,
        repository: Repository,
    ) -> SafeString:
        """Return instructions for setting up incoming WebHooks.

        Args:
            request (django.http.HttpRequest):
                The request from the client.

            repository (reviewboard.scmtools.models.Repository):
                The repository to return instructions for.

        Returns:
            django.utils.safestring.SafeString:
            The HTML to send to the client for hook instructions.
        """
        repository_owner = urlquote(repository.extra_data['repository_owner'])
        repository_name = urlquote(repository.extra_data['repository_name'])

        add_webhook_url = urljoin(
            self.account.hosting_url,
            f'{repository_owner}/{repository_name}/settings/hooks/forgejo/new')

        webhook_endpoint_url = build_server_url(local_site_reverse(
            'forgejo-hooks-close-submitted',
            local_site=repository.local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': repository.hosting_account.service_name,
            },
        ))

        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        return render_to_string(
            template_name='hostingsvcs/forgejo/repo_hook_instructions.html',
            request=request,
            context={
                'example_id': example_id,
                'example_url': example_url,
                'repository': repository,
                'server_url': get_server_url(),
                'add_webhook_url': add_webhook_url,
                'webhook_endpoint_url': webhook_endpoint_url,
                'hook_uuid': repository.get_or_create_hooks_uuid(),
            })

    def _create_api_token_name(
        self,
        server_name: str,
    ) -> str:
        """Return a name to use for a new API token.

        Args:
            server_name (str):
                The hostname of the Review Board server.

        Returns:
            str:
            A name to use for the new API token.
        """
        id = uuid.uuid4().hex[:6]

        # API token names are limited to 255 chars. It's pretty unlikely that
        # someone's server name will be this long, but just in case:
        server_name = server_name[:236]

        return f'reviewboard-{server_name}-{id}'



"""Resource representing the commits in a multi-commit review request."""

from urllib.parse import urlencode

from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from djblets.util.decorators import augment_method_from
from djblets.util.http import get_http_requested_mimetype, set_last_modified
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import (DateTimeFieldType, DictFieldType,
                                   IntFieldType, StringFieldType)

from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.diffviewer.validators import COMMIT_ID_LENGTH
from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_local_site,
                                           webapi_check_login_required)
from reviewboard.webapi.resources import resources


class DiffCommitResource(WebAPIResource):
    """Provides information on a collection of commits in a review request.

    Each diff commit resource contains individual per-file diffs as child
    resources, as well as metadata to reproduce the actual commits in a
    version control system.
    """

    added_in = '4.0'
    model = DiffCommit
    name = 'commit'
    uri_template_name = 'diff_commit'

    model_parent_key = 'diffset'
    model_object_key = 'commit_id'

    uri_object_key = 'commit_id'
    uri_object_key_regex = r'[A-Za-z0-9]{1,%s}' % COMMIT_ID_LENGTH

    allowed_methods = ('GET', 'PUT')

    required_features = [dvcs_feature]

    allowed_mimetypes = WebAPIResource.allowed_mimetypes + [
        {'item': 'text/x-patch'},
    ]

    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the commit resource.',
        },
        'author_name': {
            'type': StringFieldType,
            'description': 'The name of the author of this commit.',
        },
        'author_date': {
            'type': DateTimeFieldType,
            'description': 'The date and time this commit was authored in ISO '
                           '8601 format (YYYY-MM-DD HH:MM:SS+ZZZZ).',
        },
        'author_email': {
            'type': StringFieldType,
            'description': 'The e-mail address of the author of this commit.',
        },
        'commit_id': {
            'type': StringFieldType,
            'description': 'The ID of this commit.',
        },
        'committer_name': {
            'type': StringFieldType,
            'description': 'The name of the the committer of this commit, if '
                           'applicable.',
        },
        'committer_date': {
            'type': StringFieldType,
            'description': 'The date and time this commit was committed in '
                           'ISO 8601 format (YYYY-MM-DD HH:MM:SS+ZZZZ).',
        },
        'committer_email': {
            'type': StringFieldType,
            'description': 'The e-mail address of the committer of this '
                           'commit.',
        },
        'commit_message': {
            'type': StringFieldType,
            'description': 'The commit message.',
        },
        'extra_data': {
            'type': DictFieldType,
            'description': 'Extra data as part of the commit. This can be set '
                           'by the API or extensions.',
        },
        'filename': {
            'type': StringFieldType,
            'description': 'The name of the corresponding diff.',
        },
        'parent_id': {
            'type': StringFieldType,
            'description': 'The ID of the parent commit.',
        },
    }

    def get_queryset(self, request, *args, **kwargs):
        """Return the queryset for the available commits.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            The queryset of all available commits.
        """
        try:
            diffset = resources.diff.get_object(request, *args, **kwargs)
        except DiffSet.DoesNotExist:
            return self.model.objects.none()

        return self.model.objects.filter(diffset=diffset)

    def has_access_permissions(self, request, commit, *args, **kwargs):
        """Return whether or not the user has access permissions to the commit.

        A user has access permissions for a commit if they have permission to
        access the parent review request.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to access the commit.
        """
        review_request = resources.review_request.get_object(request, *args,
                                                             **kwargs)
        return review_request.is_accessible_by(request.user)

    def has_list_access_permissions(self, request, *args, **kwargs):
        """Return whether the user has access permissions to the list resource.

        A user has list access permissions if they have permission to access
        the parent review request.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to access the list resource.
        """
        review_request = resources.review_request.get_object(request, *args,
                                                             **kwargs)
        return review_request.is_accessible_by(request.user)

    def has_modify_permissions(self, request, obj, *args, **kwargs):
        """Return whether the user has access permissions to modify the object.

        A user has modify permissions for a commit if they have permission to
        modify the parent review request.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to modify the object.
        """
        review_request = resources.review_request.get_object(request, *args,
                                                             **kwargs)
        return review_request.is_mutable_by(request.user)

    @webapi_response_errors(DOES_NOT_EXIST)
    @augment_method_from(WebAPIResource)
    def get_list(self, request, *args, **kwargs):
        """Return the list of commits."""

    @webapi_check_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST)
    def get(self, request, *args, **kwargs):
        """Return information about a commit.

        If the :mimetype:`text/x-patch` mimetype is requested, the contents of
        the patch will be returned.

        Otherwise, metadata about the commit (such as author name, author date,
        etc.) will be returned.
        """
        mimetype = get_http_requested_mimetype(
            request,
            [mimetype['item'] for mimetype in self.allowed_mimetypes])

        if mimetype != 'text/x-patch':
            return super(DiffCommitResource, self).get(request, *args,
                                                       **kwargs)

        try:
            review_request = resources.review_request.get_object(
                request, *args, **kwargs)
            commit = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_access_permissions(request, commit, *args, **kwargs):
            return self.get_no_access_error(request)

        tool = review_request.repository.get_scmtool()
        data = tool.get_parser(b'').raw_diff(commit)

        rsp = HttpResponse(data, content_type=mimetype)
        rsp['Content-Disposition'] = ('inline; filename=%s.patch'
                                      % commit.commit_id)

        set_last_modified(rsp, commit.last_modified)
        return rsp

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    @webapi_request_fields(allow_unknown=True)
    def update(self, request, extra_fields=None, *args, **kwargs):
        """Update a commit.

        This is used solely for modifying the extra data on a commit. The
        contents of a commit cannot be modified.

        Extra data can be stored for later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            commit = self.get_object(request, *args, **kwargs)
        except DiffCommit.DoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_modify_permissions(request, commit, *args, **kwargs):
            return self.get_no_access_error(request)

        if extra_fields:
            try:
                self.import_extra_data(commit, commit.extra_data, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

            commit.save(update_fields=['extra_data'])

        return 200, {
            self.item_result_key: commit,
        }

    def _get_files_link(self, commit, request, diff_resource,
                        filediff_resource, files_key, *args, **kwargs):
        """Return the link for the files.

        As an alternative to having a per-commit files resources (and draft
        resource equivalent), this method generates link to the
        :py:class:`~reviewboard.webapi.resources.filediff.FileDiffResource`
        (or :py:class:`~reviewboard.webapi.resources.draft_filediff.
        DraftFileDiffResource` for commits on a review request draft) filtered
        specifically for this commit.

        Args:
            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The commit to retrieve the link for.

            request (django.http.HttpRequest):
                The current HTTP request.

            diff_resource (reviewboard.webapi.resources.diff.DiffResource):
                Either the diff resource (if this is the diff commit resource)
                or the draft diff resource (if this is the draft diff commit
                resource).

            filediff_resource (reviewboard.webapi.resources.filediff.
                               FileDiffResource):
                Either the filedif resource (if this is the diff commit
                resource) or the draft filediff resource (if this is the draft
                diff commit resource).

            files_key (unicode):
                The key that maps to the files link in the ``links`` field of
                the ``diff_resource``.

            *args (tuple):
                Additional positional argument.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The link information for the file resource that contains only the
            :py:class:`FileDiffs
            <reviewboard.diffviewer.models.filediff.FileDiff>` of the
            requested commit.
        """
        files_link = diff_resource.get_links(
            [filediff_resource],
            commit.diffset,
            request,
            *args,
            **kwargs)[files_key]

        files_link['href'] = '%s?%s' % (
            files_link['href'],
            urlencode({'commit-id': commit.commit_id}),
        )

        return files_link

    def get_related_links(self, obj=None, request=None, *args, **kwargs):
        """Return the related links for the resource.

        If this is for an item resource, this will return links for all the
        associated
        :py:class:`~reviewboard.diffviewer.models.filediff.FileDiffs`.

        Args:
            obj (reviewboard.diffviewer.models.diffcommit.DiffCommit,
                 optional):
                The DiffCommit to get links for.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The related links.
        """
        links = {}

        if obj is not None and request:
            links['files'] = self._get_files_link(
                commit=obj,
                request=request,
                diff_resource=resources.diff,
                filediff_resource=resources.filediff,
                files_key='files',
                *args,
                **kwargs)

        return links


diffcommit_resource = DiffCommitResource()



"""Unit tests for the GitHub hosting service."""

from __future__ import annotations

import hashlib
import hmac
from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.test.client import RequestFactory
from django.utils.safestring import SafeText
from djblets.testing.decorators import add_fixtures

from reviewboard.scmtools.core import Branch, Commit
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            RepositoryError)
from reviewboard.hostingsvcs.github import GitHub, _is_fine_grained_pat

from reviewboard.hostingsvcs.hook_utils import logger
from reviewboard.hostingsvcs.repository import RemoteRepository
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import SCMError
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any

    from django.test.client import _MonkeyPatchedWSGIResponse

    from reviewboard.hostingsvcs.testing.testcases import HttpTestPath


class GitHubTestCase(HostingServiceTestCase[GitHub]):
    """Base class for GitHub test suites."""

    service_name = 'github'

    default_account_data = {
        'personal_token': encrypt_password('abc123'),
    }

    default_repository_extra_data = {
        'repository_plan': 'public',
        'github_public_repo_name': 'myrepo',
    }


class GitHubTests(GitHubTestCase):
    """Unit tests for the GitHub hosting service."""

    def test_service_support(self) -> None:
        """Testing GitHub service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)
        self.assertFalse(self.service_class.supports_ssh_key_association)

    def test_get_repository_fields_with_public_plan(self) -> None:
        """Testing GitHub.get_repository_fields with the public plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='public',
                fields={
                    'github_public_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git://github.com/myuser/myrepo.git',
                'mirror_path': 'git@github.com:myuser/myrepo.git',
            })

    def test_get_repository_fields_with_public_org_plan(self) -> None:
        """Testing GitHub.get_repository_fields with the public-org plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='public-org',
                fields={
                    'github_public_org_repo_name': 'myrepo',
                    'github_public_org_name': 'myorg',
                }
            ),
            {
                'path': 'git://github.com/myorg/myrepo.git',
                'mirror_path': 'git@github.com:myorg/myrepo.git',
            })

    def test_get_repository_fields_with_private_plan(self) -> None:
        """Testing GitHub.get_repository_fields with the private plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='private',
                fields={
                    'github_private_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git@github.com:myuser/myrepo.git',
                'mirror_path': '',
            })

    def test_get_repository_fields_with_private_org_plan(self) -> None:
        """Testing GitHub.get_repository_fields with the private-org plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='private-org',
                fields={
                    'github_private_org_repo_name': 'myrepo',
                    'github_private_org_name': 'myorg',
                }
            ),
            {
                'path': 'git@github.com:myorg/myrepo.git',
                'mirror_path': '',
            })

    def test_get_repo_api_url_with_public_plan(self) -> None:
        """Testing GitHub._get_repo_api_url with the public plan"""
        url = self._get_repo_api_url('public', {
            'github_public_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myuser/testrepo')

    def test_get_repo_api_url_with_public_org_plan(self) -> None:
        """Testing GitHub._get_repo_api_url with the public-org plan"""
        url = self._get_repo_api_url('public-org', {
            'github_public_org_name': 'myorg',
            'github_public_org_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myorg/testrepo')

    def test_get_repo_api_url_with_private_plan(self) -> None:
        """Testing GitHub._get_repo_api_url with the private plan"""
        url = self._get_repo_api_url('private', {
            'github_private_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myuser/testrepo')

    def test_get_repo_api_url_with_private_org_plan(self) -> None:
        """Testing GitHub._get_repo_api_url with the private-org plan"""
        url = self._get_repo_api_url('private-org', {
            'github_private_org_name': 'myorg',
            'github_private_org_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myorg/testrepo')

    def test_get_bug_tracker_field_with_public_plan(self) -> None:
        """Testing GitHub.get_bug_tracker_field with the public plan"""
        self.assertTrue(
            self.service_class.get_bug_tracker_requires_username('public'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('public', {
                'github_public_repo_name': 'myrepo',
                'hosting_account_username': 'myuser',
            }),
            'http://github.com/myuser/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_public_org_plan(self) -> None:
        """Testing GitHub.get_bug_tracker_field with the public-org plan"""
        self.assertFalse(
            self.service_class.get_bug_tracker_requires_username('public-org'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('public-org', {
                'github_public_org_name': 'myorg',
                'github_public_org_repo_name': 'myrepo',
            }),
            'http://github.com/myorg/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_private_plan(self) -> None:
        """Testing GitHub.get_bug_tracker_field with the private plan"""
        self.assertTrue(
            self.service_class.get_bug_tracker_requires_username('private'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('private', {
                'github_private_repo_name': 'myrepo',
                'hosting_account_username': 'myuser',
            }),
            'http://github.com/myuser/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_private_org_plan(self) -> None:
        """Testing GitHub.get_bug_tracker_field with the private-org plan"""
        self.assertFalse(self.service_class.get_bug_tracker_requires_username(
            'private-org'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('private-org', {
                'github_private_org_name': 'myorg',
                'github_private_org_repo_name': 'myrepo',
            }),
            'http://github.com/myorg/myrepo/issues#issue/%s')

    def test_get_repository_hook_instructions(self) -> None:
        """Testing GitHub.get_repository_hook_instructions"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        hooks_uuid = repository.get_or_create_hooks_uuid()

        request = RequestFactory().get(path='/')
        request.user = User.objects.create(username='test-user')

        content = repository.hosting_service.get_repository_hook_instructions(
            request=request,
            repository=repository)

        self.assertIsInstance(content, SafeText)
        self.assertIn(
            'https://github.com/myuser/myrepo/settings/hooks/new',
            content)
        self.assertIn(
            'http://example.com/repos/1/github/hooks/close-submitted/',
            content)
        self.assertIn(
            '<code>%s</code>' % hooks_uuid,
            content)
        self.assertIn('Review Board supports closing', content)

    def test_check_repository_public(self) -> None:
        """Testing GitHub.check_repository with public repository"""
        self._test_check_repository(plan='public',
                                    github_public_repo_name='myrepo')

    def test_check_repository_private(self) -> None:
        """Testing GitHub.check_repository with private repository"""
        self._test_check_repository(plan='private',
                                    github_private_repo_name='myrepo')

    def test_check_repository_public_org(self) -> None:
        """Testing GitHub.check_repository with public org repository"""
        self._test_check_repository(plan='public-org',
                                    github_public_org_name='myorg',
                                    github_public_org_repo_name='myrepo',
                                    expected_owner='myorg')

    def test_check_repository_private_org(self) -> None:
        """Testing GitHub.check_repository with private org repository"""
        self._test_check_repository(plan='private-org',
                                    github_private_org_name='myorg',
                                    github_private_org_repo_name='myrepo',
                                    expected_owner='myorg')

    def test_check_repository_public_not_found(self) -> None:
        """Testing GitHub.check_repository with not found error and public
        repository
        """
        self._test_check_repository_error(
            plan='public',
            github_public_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='A repository with this name was not found, '
                           'or your user may not own it.')

    def test_check_repository_private_not_found(self) -> None:
        """Testing GitHub.check_repository with not found error and private
        repository
        """
        self._test_check_repository_error(
            plan='private',
            github_private_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='A repository with this name was not found, '
                           'or your user may not own it.')

    def test_check_repository_public_org_not_found(self) -> None:
        """Testing GitHub.check_repository with not found error and
        public organization repository
        """
        self._test_check_repository_error(
            plan='public-org',
            github_public_org_name='myorg',
            github_public_org_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='A repository with this organization or name '
                           'was not found.')

    def test_check_repository_private_org_not_found(self) -> None:
        """Testing GitHub.check_repository with not found error and
        private organization repository
        """
        self._test_check_repository_error(
            plan='private-org',
            github_private_org_name='myorg',
            github_private_org_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='A repository with this organization or name '
                           'was not found, or your user may not have access '
                           'to it.')

    def test_check_repository_public_plan_private_repo(self) -> None:
        """Testing GitHub.check_repository with public plan and
        private repository
        """
        self._test_check_repository_error(
            plan='public',
            github_public_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": true}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='This is a private repository, but you have '
                           'selected a public plan.')

    def test_check_repository_private_plan_public_repo(self) -> None:
        """Testing GitHub.check_repository with private plan and
        public repository
        """
        self._test_check_repository_error(
            plan='private',
            github_private_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": false}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='This is a public repository, but you have '
                           'selected a private plan.')

    def test_check_repository_public_org_plan_private_repo(self) -> None:
        """Testing GitHub.check_repository with public organization plan and
        private repository
        """
        self._test_check_repository_error(
            plan='public-org',
            github_public_org_name='myorg',
            github_public_org_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": true}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='This is a private repository, but you have '
                           'selected a public plan.')

    def test_check_repository_private_org_plan_public_repo(self) -> None:
        """Testing GitHub.check_repository with private organization plan and
        public repository
        """
        self._test_check_repository_error(
            plan='private-org',
            github_private_org_name='myorg',
            github_private_org_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": false}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='This is a public repository, but you have '
                           'selected a private plan.')

    def test_check_repository_contents_forbidden(self) -> None:
        """Testing GitHub.check_repository with a token that can read
        repository metadata but not its contents
        """
        paths: dict[str | None, HttpTestPath] = {
            '/repos/myuser/myrepo': {
                'payload': b'{}',
            },
            '/repos/myuser/myrepo/branches': {
                'status_code': 403,
                'payload': b'{"message": "Resource not accessible '
                           b'by personal access token"}',
            },
        }

        message = (
            'Your token can access this repository\'s metadata but cannot '
            'read its contents. For fine-grained Personal Access Tokens, '
            'ensure "Contents: Read" is granted on this repository.'
        )

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=2) as ctx:
            with self.assertRaisesMessage(RepositoryError, message):
                ctx.service.check_repository(
                    plan='public',
                    github_public_repo_name='myrepo')

    def test_is_fine_grained_pat(self) -> None:
        """Testing _is_fine_grained_pat token-prefix detection"""
        self.assertTrue(_is_fine_grained_pat('github_pat_' + 'a' * 70))
        self.assertFalse(_is_fine_grained_pat('ghp_' + 'a' * 36))
        self.assertFalse(_is_fine_grained_pat('a' * 40))
        self.assertFalse(_is_fine_grained_pat(''))

    def test_authorize_with_classic_pat(self) -> None:
        """Testing GitHub.authorize with a classic Personal Access Token"""
        token = 'ghp_' + 'a' * 36
        paths: dict[str | None, HttpTestPath] = {
            '/user': {
                'headers': {
                    'X-OAuth-Scopes': 'repo',
                },
                'payload': b'{}',
            },
        }

        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(
                username='myuser',
                password=token)

        self.assertTrue(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user',
            method='GET',
            username='myuser',
            password=token)

        self.assertIn('personal_token', hosting_account.data)
        self.assertNotIn('authorizations', hosting_account.data)
        self.assertEqual(
            decrypt_password(hosting_account.data['personal_token']),
            token)

    def test_authorize_with_classic_pat_missing_scopes(self) -> None:
        """Testing GitHub.authorize with a classic Personal Access Token
        missing required scopes
        """
        token = 'ghp_' + 'a' * 36
        paths = {
            '/user': {
                'headers': {
                    'X-OAuth-Scopes': 'user',
                },
                'payload': b'{}',
            },
        }

        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        message = (
            'This GitHub classic Personal Access Token must have the '
            'following scopes enabled: repo'
        )

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(AuthorizationError, message):
                ctx.service.authorize(
                    username='myuser',
                    password=token)

        self.assertFalse(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user',
            method='GET',
            username='myuser',
            password=token)

        self.assertNotIn('personal_token', hosting_account.data)
        self.assertNotIn('authorizations', hosting_account.data)

    def test_authorize_with_fine_grained_pat(self) -> None:
        """Testing GitHub.authorize with a fine-grained Personal Access Token
        """
        token = 'github_pat_' + 'a' * 70
        paths: dict[str | None, HttpTestPath] = {
            '/user': {
                'payload': b'{}',
            },
        }

        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(
                username='myuser',
                password=token)

        self.assertTrue(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user',
            method='GET',
            username='myuser',
            password=token)

        self.assertIn('personal_token', hosting_account.data)
        self.assertEqual(
            decrypt_password(hosting_account.data['personal_token']),
            token)

    def test_authorize_with_fine_grained_pat_skips_scope_check(self) -> None:
        """Testing GitHub.authorize with a fine-grained Personal Access Token
        skips the scope-header check even when the header is missing
        """
        token = 'github_pat_' + 'a' * 70
        # Deliberately omit X-OAuth-Scopes, because requests made with
        # fine-grained PATs don't return it.
        paths: dict[str | None, HttpTestPath] = {
            '/user': {
                'payload': b'{}',
            },
        }

        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(
                username='myuser',
                password=token)

        self.assertTrue(hosting_account.is_authorized)

    def test_authorize_with_legacy_unprefixed_token(self) -> None:
        """Testing GitHub.authorize with a legacy unprefixed token treats
        it as classic
        """
        token = 'a' * 40
        paths: dict[str | None, HttpTestPath] = {
            '/user': {
                'headers': {
                    'X-OAuth-Scopes': 'repo',
                },
                'payload': b'{}',
            },
        }

        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(
                username='myuser',
                password=token)

        self.assertTrue(hosting_account.is_authorized)

    def test_is_authorized_with_personal_token(self) -> None:
        """Testing GitHub.is_authorized with personal access token"""
        hosting_account = self.create_hosting_account(data={
            'personal_token': encrypt_password('abc123'),
        })
        self.assertTrue(hosting_account.is_authorized)

    def test_is_authorized_with_legacy_authorization(self) -> None:
        """Testing GitHub.is_authorized with legacy authorization token"""
        hosting_account = self.create_hosting_account(data={
            'authorization': {
                'token': 'abc123',
            },
        })
        self.assertTrue(hosting_account.is_authorized)

    def test_is_authorized_without_tokens(self) -> None:
        """Testing GitHub.is_authorized with legacy authorization token"""
        hosting_account = self.create_hosting_account(data={
            'authorizations': {},
        })
        self.assertFalse(hosting_account.is_authorized)

    def test_api_with_personal_accesstoken(self) -> None:
        """Testing GitHub API access with personal access token"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.hosting_account.data = {
                'personal_token': encrypt_password('my-personal-token'),
            }
            ctx.client.http_get('https://api.github.com/user')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user',
            username='myuser',
            password='my-personal-token')

    def test_api_with_legacy_auth_token(self) -> None:
        """Testing GitHub API access with legacy auth tokens"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.hosting_account.data = {
                'authorization': {
                    'token': 'my-legacy-token',
                },
            }
            ctx.client.http_get('https://api.github.com/user')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user',
            username='myuser',
            password='my-legacy-token')

    def test_get_branches(self) -> None:
        """Testing GitHub.get_branches"""
        payload = self.dump_json([
            {
                'ref': 'refs/heads/master',
                'object': {
                    'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                }
            },
            {
                'ref': 'refs/heads/release-1.7.x',
                'object': {
                    'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                }
            },
            {
                'ref': 'refs/heads/some-component/fix',
                'object': {
                    'sha': '764015ef492c8cb1546363b45fee7ab6d1a182ee',
                }
            },
            {
                'ref': 'refs/tags/release-1.7.11',
                'object': {
                    'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                }
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/myrepo/git/refs/heads',
            username='myuser',
            password='abc123')

        self.assertEqual(
            branches,
            [
                Branch(id='master',
                       commit='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       default=True),
                Branch(id='release-1.7.x',
                       commit='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       default=False),
                Branch(id='some-component/fix',
                       commit='764015ef492c8cb1546363b45fee7ab6d1a182ee',
                       default=False),
            ])

    def test_get_branches_master_default(self) -> None:
        """Testing GitHub.get_branches master default"""
        payload = self.dump_json([
            {
                'ref': 'refs/heads/main',
                'object': {
                    'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                }
            },
            {
                'ref': 'refs/heads/master',
                'object': {
                    'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                }
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/myrepo/git/refs/heads',
            username='myuser',
            password='abc123')

        self.assertEqual(
            branches,
            [
                Branch(id='main',
                       commit='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       default=False),
                Branch(id='master',
                       commit='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       default=True),
            ])

    def test_get_branches_main_default(self) -> None:
        """Testing GitHub.get_branches main default"""
        payload = self.dump_json([
            {
                'ref': 'refs/heads/other',
                'object': {
                    'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                }
            },
            {
                'ref': 'refs/heads/main',
                'object': {
                    'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                }
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/myrepo/git/refs/heads',
            username='myuser',
            password='abc123')

        self.assertEqual(
            branches,
            [
                Branch(id='other',
                       commit='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       default=False),
                Branch(id='main',
                       commit='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       default=True),
            ])

    def test_get_branches_default_fallback(self) -> None:
        """Testing GitHub.get_branches default fallback"""
        payload = self.dump_json([
            {
                'ref': 'refs/heads/branch1',
                'object': {
                    'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                }
            },
            {
                'ref': 'refs/heads/branch2',
                'object': {
                    'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                }
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/myrepo/git/refs/heads',
            username='myuser',
            password='abc123')

        self.assertEqual(
            branches,
            [
                Branch(id='branch1',
                       commit='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       default=True),
                Branch(id='branch2',
                       commit='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       default=False),
            ])

    def test_get_commits(self) -> None:
        """Testing GitHub.get_commits"""
        payload = self.dump_json([
            {
                'commit': {
                    'author': {'name': 'Christian Hammond'},
                    'committer': {'date': '2013-06-25T23:31:22Z'},
                    'message': 'Fixed the bug number for the '
                               'blacktriangledown bug.',
                },
                'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                'parents': [
                    {'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15'}
                ],
            },
            {
                'commit': {
                    'author': {'name': 'Christian Hammond'},
                    'committer': {'date': '2013-06-25T23:30:59Z'},
                    'message': "Merge branch 'release-1.7.x'",
                },
                'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                'parents': [
                    {'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792'},
                    {'sha': '6c5f3465da5ed03dca8128bb3dd03121bd2cddb2'},
                ],
            },
            {
                'commit': {
                    'author': {'name': 'David Trowbridge'},
                    'committer': {'date': '2013-06-25T22:41:09Z'},
                    'message': 'Add DIFF_PARSE_ERROR to the '
                               'ValidateDiffResource.create error list.',
                },
                'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                'parents': [],
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            commits = ctx.service.get_commits(
                repository=repository,
                start='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?sha=859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817'),
            username='myuser',
            password='abc123')

        self.assertEqual(
            commits,
            [
                Commit(author_name='Christian Hammond',
                       date='2013-06-25T23:31:22Z',
                       id='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       message=('Fixed the bug number for the '
                                'blacktriangledown bug.'),
                       parent='92463764015ef463b4b6d1a1825fee7aeec8cb15'),
                Commit(author_name='Christian Hammond',
                       date='2013-06-25T23:30:59Z',
                       id='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       message="Merge branch 'release-1.7.x'",
                       parent='f5a35f1d8a8dcefb336a8e3211334f1f50ea7792'),
                Commit(author_name='David Trowbridge',
                       date='2013-06-25T22:41:09Z',
                       id='f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                       message=('Add DIFF_PARSE_ERROR to the '
                                'ValidateDiffResource.create error list.'),
                       parent=''),
            ])

        for commit in commits:
            self.assertIsNone(commit.diff)

    def test_get_change(self) -> None:
        """Testing GitHub.get_change"""
        commit_sha = '1c44b461cebe5874a857c51a4a13a849a4d1e52d'
        parent_sha = '44568f7d33647d286691517e6325fea5c7a21d5e'
        tree_sha = '56e25e58380daf9b4dfe35677ae6043fe1743922'

        paths: dict[str | None, HttpTestPath] = {
            '/repos/myuser/myrepo/commits': {
                'payload': self.dump_json([
                    {
                        'commit': {
                            'author': {'name': 'David Trowbridge'},
                            'committer': {'date': '2013-06-25T23:31:22Z'},
                            'message': 'Move .clearfix to defs.less',
                        },
                        'sha': commit_sha,
                        'parents': [{'sha': parent_sha}],
                    },
                ])
            },
            '/repos/myuser/myrepo/compare/%s...%s' % (parent_sha,
                                                      commit_sha): {
                'payload': self.dump_json({
                    'base_commit': {
                        'commit': {
                            'tree': {'sha': tree_sha},
                        },
                    },
                    'files': [
                        {
                            'sha': '4344b3ad41b171ea606e88e9665c34cca602affb',
                            'filename': 'reviewboard/static/rb/css/defs.less',
                            'status': 'modified',
                            'patch': (
                                '@@ -182,4 +182,6 @@\n'
                                ' }\n'
                                ' \n'
                                '+.foo {\n'
                                '+}\n'
                                ' \n'
                                ' table {'
                            ),
                        },
                        {
                            'sha': '8e3129277b018b169cb8d13771433fbcd165a17c',
                            'filename': ('reviewboard/static/rb/css/'
                                         'reviews.less'),
                            'status': 'modified',
                            'patch': (
                                '@@ -1311,6 +1311,4 @@\n'
                                ' }\n'
                                ' \n'
                                '-.bar {\n'
                                '-}\n'
                                ' \n'
                                ' h1 {'
                            ),
                        },
                        {
                            'sha': '17ba0791499db908433b80f37c5fbc89b870084b',
                            'filename': 'new_filename',
                            'previous_filename': 'old_filename',
                            'status': 'renamed',
                            'patch': (
                                '@@ -1,1 +1,1 @@\n'
                                '- foo\n'
                                '+ bar'
                            ),
                        },
                    ],
                }),
            },
            '/repos/myuser/myrepo/git/trees/%s' % tree_sha: {
                'payload': self.dump_json({
                    'tree': [
                        {
                            'path': 'reviewboard/static/rb/css/defs.less',
                            'sha': '830a40c3197223c6a0abb3355ea48891a1857bfd',
                        },
                        {
                            'path': 'reviewboard/static/rb/css/reviews.less',
                            'sha': '535cd2c4211038d1bb8ab6beaed504e0db9d7e62',
                        },
                        {
                            'path': 'old_filename',
                            'sha': '356a192b7913b04c54574d18c28d46e6395428ab',
                        }
                    ],
                }),
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=3) as ctx:
            repository = ctx.create_repository()
            change = ctx.service.get_change(repository=repository,
                                            revision=commit_sha)

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?sha=1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
            username='myuser',
            password='abc123')

        ctx.assertHTTPCall(
            1,
            url=('https://api.github.com/repos/myuser/myrepo/compare/'
                 '44568f7d33647d286691517e6325fea5c7a21d5e...'
                 '1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
            username='myuser',
            password='abc123')

        ctx.assertHTTPCall(
            2,
            url=('https://api.github.com/repos/myuser/myrepo/git/trees/'
                 '56e25e58380daf9b4dfe35677ae6043fe1743922?recursive=1'),
            username='myuser',
            password='abc123')

        self.assertEqual(
            change,
            Commit(author_name='David Trowbridge',
                   date='2013-06-25T23:31:22Z',
                   id='1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                   message='Move .clearfix to defs.less',
                   parent='44568f7d33647d286691517e6325fea5c7a21d5e'))
        self.assertEqual(
            change.diff,
            b'diff --git a/reviewboard/static/rb/css/defs.less'
            b' b/reviewboard/static/rb/css/defs.less\n'
            b'index 830a40c3197223c6a0abb3355ea48891a1857bfd'
            b'..4344b3ad41b171ea606e88e9665c34cca602affb 100644\n'
            b'--- a/reviewboard/static/rb/css/defs.less\n'
            b'+++ b/reviewboard/static/rb/css/defs.less\n'
            b'@@ -182,4 +182,6 @@\n'
            b' }\n'
            b' \n'
            b'+.foo {\n'
            b'+}\n'
            b' \n'
            b' table {\n'
            b'diff --git a/reviewboard/static/rb/css/reviews.less'
            b' b/reviewboard/static/rb/css/reviews.less\n'
            b'index 535cd2c4211038d1bb8ab6beaed504e0db9d7e62'
            b'..8e3129277b018b169cb8d13771433fbcd165a17c 100644\n'
            b'--- a/reviewboard/static/rb/css/reviews.less\n'
            b'+++ b/reviewboard/static/rb/css/reviews.less\n'
            b'@@ -1311,6 +1311,4 @@\n'
            b' }\n'
            b' \n'
            b'-.bar {\n'
            b'-}\n'
            b' \n'
            b' h1 {\n'
            b'diff --git a/new_filename b/new_filename\n'
            b'rename from old_filename\n'
            b'rename to new_filename\n'
            b'index 356a192b7913b04c54574d18c28d46e6395428ab'
            b'..17ba0791499db908433b80f37c5fbc89b870084b\n'
            b'--- a/old_filename\n'
            b'+++ b/new_filename\n'
            b'@@ -1,1 +1,1 @@\n'
            b'- foo\n'
            b'+ bar\n')

    def test_get_change_with_not_found(self) -> None:
        """Testing GitHub.get_change with commit not found"""
        with self.setup_http_test(status_code=404,
                                  payload=b'{"message": "Not Found"}',
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(SCMError, 'Not Found'):
                repository = ctx.create_repository()
                ctx.service.get_change(
                    repository=repository,
                    revision='1c44b461cebe5874a857c51a4a13a849a4d1e52d')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?sha=1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
            username='myuser',
            password='abc123')

    def test_get_remote_repositories_with_owner(self) -> None:
        """Testing GitHub.get_remote_repositories with requesting
        authenticated user's repositories
        """
        base_url = 'https://api.github.com/user/repos'
        paths: dict[str | None, HttpTestPath] = {
            '/user/repos': {
                'payload': self.dump_json([
                    {
                        'id': 1,
                        'owner': {
                            'login': 'myuser',
                        },
                        'name': 'myrepo',
                        'clone_url': 'myrepo_path',
                        'mirror_url': 'myrepo_mirror',
                        'private': 'false'
                    },
                ]),
                'headers': {
                    'Link': f'<{base_url}?page=2>; rel="next"',
                },
            },
            '/user/repos?page=2': {
                'payload': self.dump_json([
                    {
                        'id': 2,
                        'owner': {
                            'login': 'myuser',
                        },
                        'name': 'myrepo2',
                        'clone_url': 'myrepo_path2',
                        'mirror_url': 'myrepo_mirror2',
                        'private': 'true'
                    },
                ]),
                'headers': {
                    'Link': f'<{base_url}?page=1>; rel="prev"',
                },
            },
        }

        # Fetch and check the first page.
        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('myuser')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user/repos',
            username='myuser',
            password='abc123')

        self.assertEqual(len(paginator.page_data), 1)
        self.assertFalse(paginator.has_prev)
        self.assertTrue(paginator.has_next)
        repo = paginator.page_data[0]

        self.assertIsInstance(repo, RemoteRepository)
        self.assertEqual(repo.id, 'myuser/myrepo')
        self.assertEqual(repo.owner, 'myuser')
        self.assertEqual(repo.name, 'myrepo')
        self.assertEqual(repo.scm_type, 'Git')
        self.assertEqual(repo.path, 'myrepo_path')
        self.assertEqual(repo.mirror_path, 'myrepo_mirror')

        # Fetch and check the second page.
        paginator.next()

        ctx.assertHTTPCall(
            1,
            url='https://api.github.com/user/repos?page=2',
            username='myuser',
            password='abc123')

        self.assertEqual(len(paginator.page_data), 1)
        self.assertTrue(paginator.has_prev)
        self.assertFalse(paginator.has_next)
        repo = paginator.page_data[0]

        self.assertIsInstance(repo, RemoteRepository)
        self.assertEqual(repo.id, 'myuser/myrepo2')
        self.assertEqual(repo.owner, 'myuser')
        self.assertEqual(repo.name, 'myrepo2')
        self.assertEqual(repo.scm_type, 'Git')
        self.assertEqual(repo.path, 'myrepo_path2')
        self.assertEqual(repo.mirror_path, 'myrepo_mirror2')

    def test_get_remote_repositories_with_other_user(self) -> None:
        """Testing GitHub.get_remote_repositories with requesting user's
        repositories
        """
        repos1 = self.dump_json([
            {
                'id': 1,
                'owner': {
                    'login': 'other',
                },
                'name': 'myrepo',
                'clone_url': 'myrepo_path',
                'mirror_url': 'myrepo_mirror',
                'private': 'false'
            }
        ])

        headers = {
            'Link': '<https://api.github.com/users/other/repos'
                    '?page=2>; rel="next"',
        }

        with self.setup_http_test(payload=repos1,
                                  headers=headers,
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('other')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/users/other/repos',
            username='myuser',
            password='abc123')

        self.assertEqual(len(paginator.page_data), 1)
        public_repo = paginator.page_data[0]
        self.assertIsInstance(public_repo, RemoteRepository)
        self.assertEqual(public_repo.id, 'other/myrepo')
        self.assertEqual(public_repo.owner, 'other')
        self.assertEqual(public_repo.name, 'myrepo')
        self.assertEqual(public_repo.scm_type, 'Git')
        self.assertEqual(public_repo.path, 'myrepo_path')
        self.assertEqual(public_repo.mirror_path, 'myrepo_mirror')

    def test_get_remote_repositories_with_org(self) -> None:
        """Testing GitHub.get_remote_repositories with requesting
        organization's repositories
        """
        payload = self.dump_json([
            {
                'id': 1,
                'owner': {
                    'login': 'myorg',
                },
                'name': 'myrepo',
                'clone_url': 'myrepo_path',
                'mirror_url': 'myrepo_mirror',
                'private': 'false'
            },
            {
                'id': 2,
                'owner': {
                    'login': 'myuser',
                },
                'name': 'myrepo2',
                'clone_url': 'myrepo_path2',
                'mirror_url': 'myrepo_mirror2',
                'private': 'true'
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('myorg',
                                                            'organization')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/orgs/myorg/repos',
            username='myuser',
            password='abc123')

        self.assertEqual(len(paginator.page_data), 2)
        public_repo, private_repo = paginator.page_data

        self.assertIsInstance(public_repo, RemoteRepository)
        self.assertEqual(public_repo.id, 'myorg/myrepo')
        self.assertEqual(public_repo.owner, 'myorg')
        self.assertEqual(public_repo.name, 'myrepo')
        self.assertEqual(public_repo.scm_type, 'Git')
        self.assertEqual(public_repo.path, 'myrepo_path')
        self.assertEqual(public_repo.mirror_path, 'myrepo_mirror')

        self.assertIsInstance(private_repo, RemoteRepository)
        self.assertEqual(private_repo.id, 'myuser/myrepo2')
        self.assertEqual(private_repo.owner, 'myuser')
        self.assertEqual(private_repo.name, 'myrepo2')
        self.assertEqual(private_repo.scm_type, 'Git')
        self.assertEqual(private_repo.path, 'myrepo_path2')
        self.assertEqual(private_repo.mirror_path, 'myrepo_mirror2')

    def test_get_remote_repositories_with_defaults(self) -> None:
        """Testing GitHub.get_remote_repositories with default values"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.service.get_remote_repositories()

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user/repos',
            username='myuser',
            password='abc123')

    def test_get_remote_repositories_with_filter(self) -> None:
        """Testing GitHub.get_remote_repositories with ?filter-type="""
        with self.setup_http_test(payload=b'[]',
                                  expected_http_calls=1) as ctx:
            ctx.service.get_remote_repositories('myuser',
                                                filter_type='private')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user/repos?type=private',
            username='myuser',
            password='abc123')

    def test_get_remote_repository(self) -> None:
        """Testing GitHub.get_remote_repository"""
        payload = self.dump_json({
            'id': 1,
            'owner': {
                'login': 'myuser',
            },
            'name': 'myrepo',
            'clone_url': 'myrepo_path',
            'mirror_url': 'myrepo_mirror',
            'private': 'false'
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            remote_repository = \
                ctx.service.get_remote_repository('myuser/myrepo')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/myrepo',
            username='myuser',
            password='abc123')

        self.assertIsInstance(remote_repository, RemoteRepository)
        self.assertEqual(remote_repository.id, 'myuser/myrepo')
        self.assertEqual(remote_repository.owner, 'myuser')
        self.assertEqual(remote_repository.name, 'myrepo')
        self.assertEqual(remote_repository.scm_type, 'Git')
        self.assertEqual(remote_repository.path, 'myrepo_path')
        self.assertEqual(remote_repository.mirror_path, 'myrepo_mirror')

    def test_get_remote_repository_invalid(self) -> None:
        """Testing GitHub.get_remote_repository with invalid repository ID"""
        with self.setup_http_test(status_code=404,
                                  payload=b'{"message": "Not Found"}',
                                  expected_http_calls=1) as ctx:
            with self.assertRaises(ObjectDoesNotExist):
                ctx.service.get_remote_repository('myuser/invalid')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/repos/myuser/invalid',
            username='myuser',
            password='abc123')

    def _test_check_repository(
        self,
        expected_owner: str = 'myuser',
        **kwargs,
    ) -> None:
        """Test checking for a repository.

        Args:
            expected_owner (str, optional):
                The expected owner of the repository.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`check_repository()
                <reviewboard.hostingsvcs.gitlab.GitLab.check_repository>`.
        """
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=2) as ctx:
            ctx.service.check_repository(**kwargs)

        repo_url = f'https://api.github.com/repos/{expected_owner}/myrepo'
        ctx.assertHTTPCall(
            0,
            url=repo_url,
            username='myuser',
            password='abc123')
        ctx.assertHTTPCall(
            1,
            url=f'{repo_url}/branches',
            username='myuser',
            password='abc123')

    def _test_check_repository_error(
        self,
        http_status: int,
        payload: bytes,
        expected_url: str,
        expected_error: str,
        **kwargs,
    ) -> None:
        """Test error conditions when checking for a repository.

        Args:
            http_status (int):
                The HTTP status to simulate returning.

            payload (bytes):
                The payload to return, if ``http_status`` is 200.

            expected_url (str):
                The expected URL accessed (minus any query strings).

            expected_error (str):
                The expected error message from a raised exception.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`check_repository()
                <reviewboard.hostingsvcs.gitlab.GitLab.check_repository>`.
        """
        if http_status != 200:
            payload = b'{"message": "not Found"}'

        with self.setup_http_test(status_code=http_status,
                                  payload=payload,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_error):
                ctx.service.check_repository(**kwargs)

        ctx.assertHTTPCall(
            0,
            url=expected_url,
            username='myuser',
            password='abc123')

    def _get_repo_api_url(
        self,
        plan: str,
        fields: Mapping[str, Any],
    ) -> str:
        """Return the base API URL for a repository.

        Args:
            plan (str):
                The name of the plan.

            fields (dict):
                Fields containing repository information.

        Returns:
            str:
            The API URL for the repository.
        """
        account = self.create_hosting_account()

        repository = self.create_repository(
            hosting_account=account,
            extra_data={
                'repository_plan': plan,
            })

        form = self.get_form(plan, fields)
        form.save(repository)

        return account.service._get_repo_api_url(repository)


class CloseSubmittedHookTests(GitHubTestCase):
    """Unit tests for the GitHub close-submitted webhook."""

    fixtures = ['test_users', 'test_scmtools']

    def test_close_submitted_hook(self) -> None:
        """Testing GitHub close_submitted hook with event=push"""
        self._test_post_commit_hook()

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_local_site(self) -> None:
        """Testing GitHub close_submitted hook with event=push and using a
        Local Site
        """
        self._test_post_commit_hook(
            LocalSite.objects.get(name=self.local_site_name))

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_unpublished_review_request(
        self,
    ) -> None:
        """Testing GitHub close_submitted hook with event=push and an
        un-published review request
        """
        self._test_post_commit_hook(publish=False)

    def test_close_submitted_hook_ping(self) -> None:
        """Testing GitHub close_submitted hook with event=ping"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid(),
            event='ping')
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_repo(self) -> None:
        """Testing GitHub close_submitted hook with event=push and invalid
        repository
        """
        repository = self.create_repository()

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_site(self) -> None:
        """Testing GitHub close_submitted hook with event=push and invalid
        Local Site
        """
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            local_site_name='badsite',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_service_id(self) -> None:
        """Testing GitHub close_submitted hook with event=push and invalid
        hosting service ID
        """
        # We'll test against Bitbucket for this test.
        account = self.create_hosting_account()
        account.service_name = 'bitbucket'
        account.save()

        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_event(self) -> None:
        """Testing GitHub close_submitted hook with invalid event"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid(),
            event='foo')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_signature(self) -> None:
        """Testing GitHub close_submitted hook with invalid signature"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url,
            secret='bad-secret')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_review_requests(self) -> None:
        """Testing GitHub close_submitted hook with event=push and invalid
        review requests
        """
        self.spy_on(logger.error)

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url='/r/9999/',
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

        self.assertTrue(logger.error.called_with(
            'close_all_review_requests: Review request #%s does not exist.',
            9999))

    def _test_post_commit_hook(
        self,
        local_site: (LocalSite | None) = None,
        publish: bool = True,
    ) -> None:
        """Testing posting to a commit hook.

        This will simulate pushing a commit and posting the resulting webhook
        payload from GitHub to the handler for the hook.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning the review request.

            publish (bool, optional):
                Whether to test with a published review request.
        """
        account = self.create_hosting_account(local_site=local_site)
        repository = self.create_repository(hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    local_site=local_site,
                                                    publish=publish)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            local_site=local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.SUBMITTED)
        self.assertEqual(review_request.changedescs.count(), 1)

        changedesc = review_request.changedescs.get()
        self.assertEqual(changedesc.text, 'Pushed to master (1c44b46)')

    def _post_commit_hook_payload(
        self,
        post_url: str,
        review_request_url: str,
        secret: str,
        event: str = 'push',
    ) -> _MonkeyPatchedWSGIResponse:
        """Post a payload for a hook for testing.

        Args:
            post_url (str):
                The URL to post to.

            review_request_url (str):
                The URL of the review request being represented in the
                payload.

            secret (str):
                The HMAC secret for the message.

            event (str, optional):
                The webhook event.

        Results:
            django.core.handlers.request.wsgi.WSGIRequest:
            The post request.
        """
        payload = self.dump_json({
            # NOTE: This payload only contains the content we make
            #       use of in the hook.
            'ref': 'refs/heads/master',
            'commits': [
                {
                    'id': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    'message': 'This is my fancy commit\n'
                               '\n'
                               'Reviewed at http://example.com%s'
                               % review_request_url
                },
            ]
        })

        m = hmac.new(secret.encode('utf-8'), payload, hashlib.sha1)

        return self.client.post(
            post_url,
            payload,
            content_type='application/json',
            HTTP_X_GITHUB_EVENT=event,
            HTTP_X_HUB_SIGNATURE='sha1=%s' % m.hexdigest())
